;  AMB2.ASM  [ For DOS ] Much help from Annie from alt.lang.asm newsgroup
;  An ambulance races across the bottom of the screen
;  with its siren wailing and its light flashing.
;
;  Code added to turn off cursor allows this to work in a DOS box under
;  Win XP.
;
;   Animation code extracted from the "Ambulance"
;   by AK   10/02
;
.model  tiny
.386
.code
org     100h                    ; DOS .COM file

begin:
        jmp     start                   ; go start the program
;
BDA_addr                equ     40h     ; BDA (Bios Data Area) segment address
BDA_video_mode          equ     49h     ; BDA location of current video mode
BDA_timer_counter       equ     6Ch     ; BDA location of nmbr of timer ticks
                                        ; (18.2 per second) since midnight
counter dw      9
;
start:
        mov     ch,10h                  ; set bits to turn cursor off
        call    cursor                  ; do it
host:
        nop
;
real_start:
        call    find_displacement
        db      01h                     ; Used to make this add up to 3 bytes
find_displacement:
        pop     si
        sub     si, offset host
        call    payload
next_step:
        ret

payload:
        push    es
        mov     ax, word ptr [si+counter-4]
        and     ax, 00000111b
        cmp     ax, 00000110b          ; Show payload every eighth
                                        ; (starting with the sixth) time
;
; Did we already show the car?
;
        mov     ax, BDA_addr
        mov     es, ax
        call    show_payload
exit_payload:
        pop     es
        mov     ch,5                    ; set bits to turn cursor on
        mov     cl,6                    ; ditto
        call    cursor                  ; go do it
        int     20h                     ; exit to DOS
;
; Setup and show the 'ambulance car' 
;
show_payload:
;
; Check video mode
;
; Text mode 3 (80x25) - video buffer address (color) = 0B800h
; Text mode 7 (80x25) - video buffer address (monochrome) = 0B000h
;
        push    ds
        mov     di, 0B800h
        mov     ax, BDA_addr
        mov     ds, ax
        mov     al, ds:BDA_video_mode
        cmp     al, 7                   ; Check which video mode we're
        jne     setup_video_n_tune      ; on, if not Monochrome text
        mov     di, 0B000h              ; mode 7, assume mode 3
 
setup_video_n_tune:
        mov     es, di
        pop     ds
        mov     bp, 0FFF0h              ; Setup number of tones to play
                                        ; (will increment up to 50h)
setup_animation:
        mov     dx, 0                   ; Setup ambulance_data column
        mov     cx, 16                  ; Number of characters that make
                                        ; up one ambulance_data line
do_ambulance:
        call    show_ambulance          ; Print the ambulance to screen
        inc     dx
        loop    do_ambulance
        call    play_siren              ; Play a tone of the 'siren'
        call    wait_tick               ; and wait for a tick
        inc     bp
        cmp     bp, 50h                 ; Already played the 'ambulance
        jne     setup_animation         ; siren' tune 12 times?
        call    speaker_off             ; If yes, then turn speaker off
        push    ds
        pop     es
        ret
;
; Turn the PC speaker off
;
; 8255 PPI - Programmable Peripheral Interface
; Port 61h, 8255 Port B output
;
; (see description below)
;
speaker_off:
        in      al, 61h
        and     al, 11111100b          ; Disable timer channel 2 and 'ungate'
        out     61h, al                 ; its output to the speaker
        ret
;
; Turn on the speaker and play the "ambulance siren" sound
;
; Select tone frequency to generate
;
; Tone frequency is selected by means of the 3rd least significant bit of BP
;
;
; Bit(s)                        Description
; ------                        -----------
; ... 3 2 1 0
; ... x 0 x x                    Play 1st tone frequency
; ... x 1 x x                    Play 2nd tone frequency
;
; If we consider A to be the 1st tone and B to be the 2nd tone,
; then the whole 'ambulance siren' tune will be: (AAAABBBB) x 12
;
play_siren:
        mov     dx, 07D0h               ; "ambulance siren" 1st tone freq.
        test    bp, 00000100b          ; Check if we are to play
        jz      speaker_on              ; the first or the second
                                        ; tone frequency
        mov     dx, 0BB8h           ; "ambulance siren" 2nd tone frequency
;
; Turn on the speaker
;
; 8255 PPI - Programmable Peripheral Interface
; Port 61h, 8255 Port B output
;
; Bit(s)                        Description
; ------                        -----------
; 7 6 5 4 3 2 1 0
; . . . . . . . 1                Timer 2 gate to speaker enable
; . . . . . . 1 .                Speaker data enable
; x x x x x x . .                Other non-concerning fields
;
speaker_on:
        in      al, 61h
        test    al, 00000011b          ; If speaker is already on, then
        jnz     play_tone               ; play the sound tone
        or      al, 00000011b          ; Else, enable timer channel 2 and
        out     61h, al                 ; 'gate' its output to the speaker
;
; Program the PIT
;
; 8253 PIT - Programmable Interval Timer
; Port 43h, 8253 Mode Control Register
;
; Bit(s)                        Description
; ------                        -----------
; 7 6 5 4 3 2 1 0
; . . . . . . . 0                16 binary counter
; . . . . 0 1 1 .                Mode 3, square wave generator
; . . 1 1 . . . .                Read/Write LSB, followed by write of MSB
; 1 0 . . . . . .                Select counter (channel) 2
;
        mov     al, 10110110b          ; Set 8253 command register
        out     43h, al                 ; for mode 3, channel 2, etc
;
; Generate a tone from the speaker
;
; 8253 PIT - Programmable Interval Timer
; Port 42h, 8253 Counter 2 Cassette and Speaker Functions
;
play_tone:
        mov     ax, dx
        out     42h, al                 ; Send LSB (Least Significant Byte)
        mov     al, ah
        out     42h, al                 ; Send MSB (Most Significant Byte)
        ret
;
; Show the 'ambulance car'
;
show_ambulance:
        push    cx
        push    dx
        lea     bx, [si+ambulance_data-4]
        add     bx, dx                  ; Setup which ambulance_data column
                                        ; we're going to print
        add     dx, bp                  ; Don't show ambulance_data columns
        or      dx, dx                  ; which aren't still visible
        js      ambulance_done
        cmp     dx, 50h                 ; Check if the column we're printing
        jae     ambulance_done          ; is past the screen limit
                                        ; If yes,  then don't print it
        mov     di, 3200                ; Point to start of screen's 64th line
        add     di, dx                  ; Point to the column we're supposed
        add     di, dx                  ; to be printing at
        sub     dx, bp                  ; Restore to initial column value
        mov     cx, 5                   ; Set it up so we're in the first line

decode_character:
        mov     ah,4d                    ; Set color attribute to red
;                                        ; 7 = white
; Decode the character
;
; It's really pretty ingenius. Each character is encoded in such a way
; that for each line beyond the first one, that character is incremented
; by one -- and for each column beyond the first, the same thing happens.
; So taking that into account, it's not difficult to understand how it
; all works, and how to decode the ambulance_data.
;
        mov     al, [bx]                ; Get the character
        sub     al, 7
        add     al, cl                  ; Account for which line we're in
        sub     al, dl                  ; Account for which column we're in
        cmp     cx, 5                   ; Are we in the first line?
        jne     print_character         ; If we are, then...
        mov     ah,1                    ; Set color attribute to hi-inten blue
        test    bp, 00000011b           ; Is this the ending tone of a AAAA or
                                        ; BBBB tune sequence?
        jz      print_character         ; If not, then go ahead and print the
                                        ; 'siren' characters
        mov     al, ' '                 ; Else, replace them with a ' ' (to
                                        ; accomplish the visual 'siren' effect
print_character:
        stosw                           ; Print the character to screen
        add     bx, 16                  ; Point to next ambulance_data line
        add     di, 158                 ; Point to next screen line
        loop    decode_character
ambulance_done:
        pop     dx
        pop     cx
        ret
;
; Wait for one tick (18.2 per second) to pass
;
wait_tick:
        push    ds
        mov     ax, BDA_addr
        mov     ds, ax
        mov     ax, ds:BDA_timer_counter ; Get ticks since midnight
check_timer:
        cmp     ax, ds:BDA_timer_counter ; Check if one tick has
        je      check_timer              ; already passed
        pop     ds
        ret
;
; Our cursor manipulation routine.
;
cursor:
        mov     ah,1                    ; function 1 - cursor control
        int     10h                     ; call ROM BIOS video services
        ret                             ; return to caller
;
; Data from here below
;
ambulance_data:
   first_line   db     22h, 23h, 24h, 25h, 26h, 27h, 28h, 29h, 66h, 87h, 3Bh
                db     2Dh, 2Eh, 2Fh, 30h, 31h
   second_line  db     23h, 0E0h, 0E1h, 0E2h, 0E3h, 0E4h, 0E5h, 0E6h, 0E7h
                db     0E7h, 0E9h, 0EAh, 0EBh, 30h, 31h, 32h
   third_line   db     24h, 0E0h, 0E1h, 0E2h, 0E3h, 0E8h, 2Ah, 0EAh, 0E7h
                db     0E8h, 0E9h, 2Fh, 30h, 6Dh, 32h, 33h
   fourth_line  db     25h, 0E1h, 0E2h, 0E3h, 0E4h, 0E5h, 0E7h, 0E7h, 0E8h
                db     0E9h, 0EAh, 0EBh, 0ECh, 0EDh, 0EEh, 0EFh
   fifth_line   db     26h, 0E6h, 0E7h, 29h, 59h, 5Ah, 2Ch, 0ECh, 0EDh, 0EEh
                db     0EFh, 0F0h, 32h, 62h, 34h, 0F4h
end   begin
